// Package climain contains high-level logic for parsing the main Miller
// command line. Package cli contains the flags table and most handlers;
// package cli is also invoked by verbs to parse their own flags.
package climain
