// Package auxents implements little helper-tools embedded within Miller, such
// as mlr hex and mlr termcvt. It also implements mlr help (on-line help), mlr
// regtest (for regressio-testing), and mlr repl (Miller's read-evaluate-print
// loop).
package auxents
