// Package parsing contains the single source file mlr.bnf, which is the
// lexical/semantic grammar file for the Miller put/filter DSL using the GOCC
// framework. Go source files are autogenerated by GOCC.
package parsing
